package neustore.base;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * Hash/LinkList-based LRU-buffer with O(1) average time performance of all
 * functions except flush (O(n)). Fails with exceptions when unable to complete
 * a request due to pinned pages.
 * 
 * @version $Id: LRUBuffer.java 57 2009-03-08 23:47:18Z erikbagg $
 * @author Erik Bagge Ottesen <erikbagg@idi.ntnu.no>
 */
public class LRUBuffer extends DBBuffer {
    /**
     * Map with the collection views ordered oldest to newest. As get() and
     * put() both update the LRU-time, it is important to not use these when
     * iterating.
     */
    private Map<DBBufferHashKey, DBBufferStoredElement> cache;

    public LRUBuffer(int bufferSize, int pageSize) {
        super(bufferSize, pageSize);
        cache = new LinkedHashMap<DBBufferHashKey, DBBufferStoredElement>(bufferSize * 3 / 4, 0.75F, true);
    }

    /**
     * Adds the given page to the buffer.
     * 
     * @param stored element to add.
     * @exception DBRuntimeException if the buffer is full and all pages are
     *            pinned.
     */
    @Override
    protected void add(DBBufferStoredElement stored) throws IOException {
        if (cache.size() >= bufferSize) {
            // cache full, remove oldest page (that is not pinned)
            for (Entry<DBBufferHashKey, DBBufferStoredElement> entry : cache.entrySet()) {
                if (entry.getValue().pinCount == 0) {
                    flushPage(entry.getValue());
                    cache.remove(entry.getKey());
                    break;
                }
            }
            // failed to remove page
            if (cache.size() >= bufferSize) {
                throw new DBRuntimeException("Buffer is full. All pages in buffer are pinned.");
            }
        }

        DBBufferHashKey key = new DBBufferHashKey(stored.file, stored.pageID);
        cache.put(key, stored);
    }

    @Override
    protected DBBufferStoredElement find(RandomAccessFile file, int pageID) {
        DBBufferHashKey key = new DBBufferHashKey(file, pageID);
        return cache.get(key);
    }


    /**
     * Flushes all pages of the given file to disk but does not remove them from the
     * buffer. If file is null, all pages are flushed to disk.
     * 
     * @param file file to flush to disk, or null if all files.
     * @exception DBRuntimeException if the any pages that are to be flushed are
     *            pinned.
     */
    public synchronized void force(RandomAccessFile file) throws IOException {
        Iterator<Entry<DBBufferHashKey, DBBufferStoredElement>> it = cache.entrySet().iterator();
        while (it.hasNext()) {
            Entry<DBBufferHashKey, DBBufferStoredElement> entry = it.next();
            if ((file == null) || file.equals(entry.getKey().file)) {
                flushPage(entry.getValue());
            }
        }
    }
    
    /**
     * Flushes all pages of the given file to disk and removes them from the
     * buffer. If file is null, all pages are flushed to disk and the buffer is
     * emptied.
     * 
     * @param file file to flush to disk, or null if all files.
     * @exception DBRuntimeException if the any pages that are to be flushed are
     *            pinned.
     */
    @Override
    public synchronized void flush(RandomAccessFile file) throws IOException {
        Iterator<Entry<DBBufferHashKey, DBBufferStoredElement>> it = cache.entrySet().iterator();
        while (it.hasNext()) {
            Entry<DBBufferHashKey, DBBufferStoredElement> entry = it.next();
            if ((file == null) || file.equals(entry.getKey().file)) {
                flushPage(entry.getValue());
                it.remove();
            }
        }
    }

    /**
     * Flushes a given element (page) to disk
     * 
     * @param stored the element to flush to disk.
     * @exception DBRuntimeException if the given page is pinned.
     */
    private void flushPage(DBBufferStoredElement stored) throws IOException {
        if (stored.pinCount > 0) {
            throw new DBRuntimeException("Page is pinned, can not flush.");
        }
        if (stored.dirty) {
            byte[] page = new byte[pageSize];
            if (stored.parsed) {
                ((DBPage) stored.object).write(page);
            } else {
                System.arraycopy((byte[]) stored.object, 0, page, 0, pageSize);
            }
            save(stored.file, page, stored.pageID);
            stored.dirty = false;
        }
    }

    /**
     * Resets the buffer size. Note that when reducing the buffer size if the
     * currently active buffer size > bufferSize, this will not have the
     * intended effect. For this to work properly a flush(null) should be
     * performed first.
     * 
     * @param bufferSize new buffer size
     */
    public void setBufferSize(int bufferSize) {
        this.bufferSize = bufferSize;
    }
}
