package neustore.base;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * A class that stores a variable-length String.
 * Note that to write a StringData to disk, the occupied size is four bytes
 * larger than the String size, for we also store the size of the String. 
 * 
 * @author Donghui Zhang &lt;donghui@ccs.neu.edu&gt;
 */
public class StringData implements Data<String> {
	/**
	 * maximum length of a String.
	 */
	protected int maxStringLength;
	/**
	 * the stored String
	 */
	private String string;
	
	/**
	 * The constructor takes as input the max length of a string.
	 * The maxSize of this object is this maxStringLength + 4.
	 * @param _maxStringLength   max length of a string
	 */
	public StringData( int _maxStringLength, String _string ) {
		maxStringLength = _maxStringLength;
		string = new String(_string);
	}
	
	public Object clone() {
		StringData newData = new StringData(maxStringLength, string);
		return newData;
	}

	public int size() {
	        byte[] bytes = string.getBytes();
	        return Math.min(maxStringLength, bytes.length) + 4;
	}

	public int maxSize() {
		return maxStringLength + 4;
	}

	public void read(DataInputStream in) throws IOException {
		int length = in.readInt();
                if (length < 0) {
                    throw new DBRuntimeException();
                }
		if ( length == 0 ) {
			string = "";
			return;
		}
		byte[] b = new byte[length];
		in.read(b);
		string = new String(b);
	}

	public void write(DataOutputStream out) throws IOException {
		byte[] b = string.getBytes();
		int length = Math.min(maxStringLength, b.length);
		out.writeInt(length);
		out.write(b, 0, length);
	}

	public String value() {
	    return string;
	}
	
	@Override
	public String toString() {
	        return getClass().getSimpleName() + "[value=" + value() + "]";
	}
}
