package ntnustore.btree;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import neustore.base.DBBuffer;
import neustore.base.DBBufferReturnElement;
import neustore.base.DBIndex;
import neustore.base.ByteArray;
import neustore.base.DBRuntimeException;
import neustore.base.Data;
import neustore.base.FloatData;
import neustore.base.Key;
import java.util.ArrayList;

import ntnustore.btree.BTreeCursor;
import ntnustore.dlaesa.Cursor;

/**
 * BTree module. It is responsible for the management of the B-tree index, with
 * root blocks, navigation through multiple levels, splits and index insertion.
 * The BTreePage module handles records in each page
 * 
 * @author Svein Erik Bratsberg &lt;sveinbra@idi.ntnu.no&gt;
 */

public class BTree<K extends Key<?>, D extends Data<?>> extends DBIndex {

    protected int rootNode;
    protected int numRecs;
    protected int numLevels;
    protected boolean debugOn;
    boolean type = false;

    /**
     * Returns the number of records.
     * 
     * @return number of records
     */
    public int numRecs() {
        return numRecs;
    }
    
    
    private K sampleKey;
    private D sampleData;

    /**
     * The BTree constructor. It takes as additional input a sample key and a
     * sample data.
     * 
     * @param _buffer the buffer
     * @param isCreate whether create (or open)
     * @param _sampleKey the sample key
     * @param _sampleData the sample data
     * @throws IOException
     */
    public BTree(DBBuffer _buffer, String filename, boolean isCreate, K _sampleKey, D _sampleData) throws IOException {
        super(_buffer, filename, isCreate);
        //System.out.println("sampledata " + _sampleData);
        sampleKey = _sampleKey;
        sampleData = _sampleData;
        debugOn = false;
        if (isCreate) { // must do it here, since initIndexHead()
            rootNode = allocate(); // does not throw IOException
            BTreePage<K, D> rootPage = new BTreePage<K, D>(BTreePage.BTREE_LEAF_TYPE, rootNode, 0, pageSize, sampleKey,
                    sampleData);
            buffer.writePage(file, rootNode, rootPage);
        }
    }

    public boolean insert(K key, Data<?> data) throws IOException {
        return insert(key, data, 0); // data is on level 0
    }

    /**
     * Inserts a new record into the b-tree. If the page is full, split it into
     * two pages.
     * 
     * @param key the key to be inserted
     * @param data the data to be inserted
     * @param level the level to insert into
     * @return whether inserted
     */
    public boolean insert(K key, Data<?> data, int level) throws IOException {
        numRecs++;
        BTreePage<K, Data<?>> myPage = findPage(key, level);
        myPage.getLock(true);
        int myPageId = myPage.pageId;
        if (myPage.isFull(key, data)) {
            // System.out.println("Splitting full page");
            // myPage.debug(false);
            int pageType = (myPage.level == 0) ? BTreePage.BTREE_LEAF_TYPE : BTreePage.BTREE_IDX_TYPE;
            // allocate new page
            int newPageId = allocate();
            //System.out.println("Allocating new page in split id="+newPageId+" level="+myPage.level);
            BTreePage<K, Data<?>> rightPage = new BTreePage<K, Data<?>>(pageType, newPageId, myPage.level, pageSize,
                    key, data);

            // link in the new page
            rightPage.prev = myPageId;
            rightPage.next = myPage.next;
            myPage.next = rightPage.pageId;
            if (rightPage.next != BTreePage.NOPAGE) {
                BTreePage<K, Data<?>> nextPage = myReadPage(rightPage.next);
                nextPage.prev = newPageId;
                buffer.writePage(file, nextPage.pageId, nextPage);
            }

            // Records are moved from the current page to the split page, only
            // if the new key is inside the interval of the current keys.
            // Otherwise the new key is inserted as the only record in the new
            // (right) page. This is a simplified method of handling record
            // redistribution when inserting sequential keys.
            K splitKey;
            if (key.compareTo(myPage.records.getKey(myPage.records.size() - 1)) > 0) {
                // put only new record in right page
                splitKey = key;
            } else {
                // move records to new page
                splitKey = myPage.moveRecords(rightPage);
            }

            // insert record
            if (key.compareTo(splitKey) < 0) {
                //System.out.println(key);
            	myPage.insert(key, data);
            } else {
            	//System.out.println(key);
                rightPage.insert(key, data);
            }

            buffer.writePage(file, myPage.pageId, myPage);
            buffer.writePage(file, rightPage.pageId, rightPage);

            // System.out.println("========After
            // split=========================");
            // myPage.debug(true);
            // rightPage.debug(true);

            // create index data
            IdxData newPagePtr = new IdxData(newPageId);

            // check if new root needs to be allocated
            if (myPageId == rootNode) {
                int rootPageId = allocate();
                //System.out.println("creating new root node="+rootPageId+" level="+(myPage.level+1));
                BTreePage<K, IdxData> rootPage = new BTreePage<K, IdxData>(BTreePage.BTREE_IDX_TYPE, rootPageId,
                        myPage.level + 1, pageSize, splitKey, newPagePtr);
                rootNode = rootPageId;
                // insert index record to leftmost page
                rootPage.leastPage = myPageId;
                buffer.writePage(file, rootNode, rootPage);
                numLevels++;
            }

            myPage.releaseLock();

            // insert index record
            insert(splitKey, newPagePtr, myPage.level + 1);

            return true;
        } else {
            myPage.insert(key, data);
            buffer.writePage(file, myPage.pageId, myPage);
            myPage.releaseLock();
            return true;
        }
    }

    /**
     * Deletes the record by the given key. After deletion, if the page is empty
     * it should be set free.
     * 
     * @param key the key of the record to be deleted
     * @return whether deleted
     * @throws IOException
     */
    public boolean delete(K key) throws IOException {
        BTreePage<K, Data<?>> myPage = findPage(key, 0); // data on level 0
        myPage.getLock(true);
        boolean retVal = myPage.delete(key);
        myPage.releaseLock();
        numRecs = numRecs -1;
        return retVal;
    }

    public BTreeCursor<K, D> scanFromTo(K startKey, K stopKey) throws IOException {
    	BTreePage<K, D> myPage = (BTreePage<K, D>) findPage(startKey, 0);
        return new BTreeCursor<K, D>(this, myPage, startKey, stopKey);
    }

    public Cursor<K, D> query(K rangeMin, K rangeMax) {
        try {
            return scanFromTo(rangeMin, rangeMax);
        } catch (IOException e) {
            throw new DBRuntimeException(e);
        }
    }
    
    /**
     * Searches the record by the given key.
     * 
     * @param key the key to search
     * @return <code>data</code> if found; <code>null</code> otherwise
     * @throws IOException
     */
    @SuppressWarnings("unchecked")
	public D search(K key) throws IOException {
        BTreePage<K, Data<?>> myPage = findPage(key, 0); // data on level 0
        //System.out.println("searchpageid "+myPage.pageId);
        myPage.getLock(false);
        D retVal = (D) myPage.search(key);
        myPage.releaseLock();
        return retVal;
    }
   
    /**
     * Navigate to the right page at the right level
     * 
     * @param key the key to search
     * @param level level in B-tree, 0=leaf
     * @return page
     * @throws IOException
     */
    protected BTreePage<K, Data<?>> findPage(K key, int level) throws IOException {
        BTreePage<K, Data<?>> navPage = myReadPage(rootNode);
        // getLock after reading the level
        while (navPage.level > level) {
            navPage.getLock(false); // don't need write lock here
            if (debugOn) {
                System.out.println("Search key=" + key.toString());
                navPage.debug(true);
            }
            int navId = navPage.searchLessEqual(key);
            //System.out.println("navid " + navId);
            navPage.releaseLock();
            navPage = myReadPage(navId);
        }
        if (debugOn) {
            System.out.println("Search key=" + key.toString());
            navPage.debug(true);
        }
        return navPage;
    }
    

    protected <T extends Data<?>> BTreePage<K, T> getNext(BTreePage<K, T> currPage) throws IOException {
        if (currPage.next != BTreePage.NOPAGE)
            return (BTreePage<K, T>) myReadPage(currPage.next);
        else
            return null;
    }

    /**
     * Reads a page from buffer and parse it if not parsed already.
     * 
     * @param pageID
     * @return an object of class BTreePage
     * @throws IOException
     */
    protected BTreePage<K, Data<?>> myReadPage(int pageID) throws IOException {
        DBBufferReturnElement ret = buffer.readPage(file, pageID);
        BTreePage<K, Data<?>> thePage = null;
        if (ret.parsed) {
            thePage = (BTreePage<K, Data<?>>) ret.object;
        } else {
            // read pageType to check whether to create index or leaf page
            byte[] page = (byte[]) ret.object;
            ByteArrayInputStream byte_in = new ByteArrayInputStream(page);
            DataInputStream in = new DataInputStream(byte_in);
            int pageType = in.readInt();
            if (pageType == BTreePage.BTREE_IDX_TYPE) {
                thePage = new BTreePage<K, Data<?>>(pageType, BTreePage.NOPAGE, -1, pageSize, sampleKey, new IdxData(
                        BTreePage.NOPAGE));
            } else {
                thePage = new BTreePage<K, Data<?>>(pageType, BTreePage.NOPAGE, -1, pageSize, sampleKey, sampleData);
            }
            thePage.read(page);
            //System.out.println("=============myReadPage=====================");
            //thePage.debug(true);
        }
        return thePage;
    }

    protected void readIndexHead(byte[] indexHead) {
        ByteArray ba = new ByteArray(indexHead, true);
        try {
            numRecs = ba.readInt();
            rootNode = ba.readInt();
            numLevels = ba.readInt();
        } catch (IOException e) {
        }
    }

    protected void writeIndexHead(byte[] indexHead) {
        ByteArray ba = new ByteArray(indexHead, false);
        try {
            ba.writeInt(numRecs);
            ba.writeInt(rootNode);
            ba.writeInt(numLevels);
        } catch (IOException e) {
        }
    }

    protected void initIndexHead() {
        numRecs = 0;
        numLevels = 1;
    }

    @Override
    public String toString() {
        return "BTree[pages=" + numPages + ", records=" + numRecs + "]";
    }

    /**
     * Return statistics, i.e. an ArrayList containing sample values at equiv-depth.
     * @param numBins    how many values should be sampled.
     */

    public ArrayList<K> createEquivDepthStatistics (int numBins) throws IOException {
        ArrayList<K> values = new ArrayList<K>();
        // Find correct level in B-tree to collect samples
        BTreePage<K, Data<?>> navPage = myReadPage(rootNode);
        int pageRecs = navPage.numRecs();
        if (pageRecs>numBins) {
            float step = (float)numBins/pageRecs;
            float stepSum = step;
            int binI = 0;
            for (int i = 0; i < pageRecs; i++) {
                if (stepSum>=binI) {
                    values.add(navPage.getKey(i));
                    binI++;
                }
                stepSum += step;
            }
        }
        return values;

    }
}
