package ntnustore.btree;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import neustore.base.DBPage;
import neustore.base.DBRuntimeException;
import neustore.base.Data;
import neustore.base.FloatData;
import neustore.base.Key;

import ntnustore.btree.SortedRecordList;

/**
 * A memory-version page of a b-tree.
 * 
 * @see BTree
 * @author Svein Erik Bratsberg &lt;sveinbra@idi.ntnu.no&gt;
 */
public class BTreePage<K extends Key<?>, D extends Data<?>> extends DBPage {
    protected int pageId; // myId
    protected int level;  // 0=leaf
    protected int next;   // next page, NOPAGE if not exist
    protected int prev;   // prev page, NOPAGE if not exist
    protected int leastPage; // least page at level below, NOPAGE if not applicable
    protected int availableBytes;

    // lock variables
    private boolean writeLocked;
    private boolean readLocked;
    private int nReaders;

    // int numRecs; is stored on disk, but not in memory

    /**
     * size for the reserved space.
     * The first bytes of the disk-version of BTreePage are reserved for:
     * nodeType, prev, next, availableBytes, numRecs, etc
     */
    protected final int RESERVED = 36;
    protected static final int BTREE_LEAF_TYPE = 2;
    protected static final int BTREE_IDX_TYPE = 3;
    protected static final int NOPAGE = -1;
	
    /**
     * sample key used in readPage
     */
    private K sampleKey;
    
    /**
     * sample data used in readPage
     */
    private D sampleData;
	
    /**
     * the list of records
     */
    protected SortedRecordList<K, D> records;

    /**
     * Constructor fo BTreePage. 
     * Here a sample key and a sample data are taken to enable the
     * generic feature of the index.
     * 
     * @param _pageId
     * @param _level
     * @param _pageSize     page size
     * @param _sampleKey    a sample key
     * @param _sampleData   a sample data
     */
    public BTreePage(int _pageType, int _pageId, int _level, int _pageSize,
		     K _sampleKey, D _sampleData)
    {
	super(_pageType, _pageSize);
	pageId = _pageId;
	level = _level;
	prev = next = NOPAGE;
	leastPage = NOPAGE;
	availableBytes = pageSize - RESERVED;
	sampleKey = _sampleKey;
        sampleData = _sampleData;
        records = new SortedRecordList<K, D>();
    }

    /**
     * Returns the number of records in the page.
     * @return number of records
     */
    public int numRecs() { 
	return records.size();
    }
    
    /**
     * Whether the page is full.
     * 
     * @return  whether full
     */
    public boolean isFull() {
        return (availableBytes < sampleKey.maxSize()+sampleData.maxSize()); 
    }
    
    /**
     * Whether the page is full. That is wheter the page has room for the given key/data-pair.
     * 
     * @return  whether full
     */
    public boolean isFull(K key, D data) {
        return (availableBytes < key.size()+data.size()); 
    }

    /**
     * Inserts a new record into this page. The caller needs to make sure that
     * the page has enough space.
     * @param key   the key part of the new record
     * @param data  the data part of the new record
     */
    public void insert(K key, D data) {
	records.insert(key,data);
	availableBytes -= key.size()+data.size();
        if (availableBytes < 0) {
            throw new DBRuntimeException();
        }
	//System.out.println("pageId="+pageId+" availableBytes="+availableBytes);
    }
	
    /**
     * Searches for a record in the page.
     * @param   key   the key to search for
     * @return  the data if found; null otherwise.
     */
    /*
    public D search(K key) {
	return records.find(key);
    }
    */
    public D search(K key) {
    	//System.out.println("PAGEID "+ this.pageId);
    	return records.find(key);
        }
	
    /**
     * @param i idx
     * @return  the key at pos i
     */
    public K getKey(int i) {
        if (i < records.size()) {
            return records.getKey(i);
        } else {
            return null;
        }
    }

    /**
     * Searches for a record in the page with the first key less or equal
     * Used when retrieving index records.
     * Returns leastPage when no lessEqual key is there, and leastPage is set.
     * @param   key   the key to search for
     * @return  the data if found; null otherwise.
     */
    public int searchLessEqual (K key) {
    //System.out.println(key);
	IdxData nxtBlock = (IdxData)records.findLeq(key);
	//System.out.println("nxtBlock " + nxtBlock);
	if (nxtBlock==null) {
            return leastPage;
	}
	return nxtBlock.value();
    }


    /**
     * Deletes a record from the page.
     * It is the caller's responsibility to free an empty page.
     * @param   key  the key to delete
     * @return  whether deleted
     */
    public boolean delete( K key ) {
	if (records.remove(key)) {
	    availableBytes += sampleKey.size() + sampleData.size();
	    return true;
	}
	return false;
    }

    protected K moveRecords (BTreePage<K, D> rightPage) throws IOException {
	// copy records to new page
	int maxIdx = numRecs();
	int splitIdx = maxIdx/2;
        //System.out.println("moveRecords splitIdx="+splitIdx+" maxIdx="+maxIdx);
	K splitKey = records.getKey(splitIdx);
        int bytesMoved = 0;
	for (int i=splitIdx;i<maxIdx;i++) {
	    rightPage.records.insertSequential(records.getKey(i),records.getData(i));
            bytesMoved += records.getKey(i).size()+records.getData(i).size();
	}
        rightPage.availableBytes -= bytesMoved;
        availableBytes += bytesMoved;
        records.removeAllFrom(splitIdx);
	return splitKey;
    }

    protected void read(byte[] page) throws IOException {
	ByteArrayInputStream byte_in = new ByteArrayInputStream(page);
	DataInputStream in = new DataInputStream(byte_in);
	in.readInt();  // skip nodeType, set in constructorer
	in.readInt();  // skip pageSize, set in constructorer
	pageId = in.readInt();
	level = in.readInt();
	next = in.readInt();
	prev = in.readInt();
	leastPage = in.readInt();
	availableBytes = in.readInt();
	int numRecs = in.readInt();
	//System.out.println("Reading pageId="+pageId+" nRecs="+numRecs);
	
	records.clear();
	for ( int i=0; i<numRecs; i++ ) {
		K rKey = (K)sampleKey.clone();
	    D rData = (D)sampleData.clone();
	    
	    //System.out.println("samplekey: " + sampleKey);
	    //System.out.println("sampledata: " + sampleData);
	    
	    rKey.read(in);
	    rData.read(in);
	   
	    //System.out.println("key " + rKey);
	    //System.out.println("Data " + rData);
	    //System.out.println("");
	    records.insertSequential(rKey, rData);
	    //System.out.println("nRecs="+records.size()+" key="+rKey.toString());
	}
	
	in.close();
	byte_in.close();
    }

    protected void write(byte[] page) throws IOException {
	ByteArrayOutputStream byte_out = new ByteArrayOutputStream(pageSize);
	DataOutputStream out = new DataOutputStream(byte_out);
	out.writeInt(nodeType);
	out.writeInt(pageSize);
	out.writeInt(pageId);
	out.writeInt(level);
	out.writeInt(next);
	out.writeInt(prev);
	out.writeInt(leastPage);
	out.writeInt(availableBytes);
	out.writeInt(records.size());

	// System.out.println("Writing pageId="+pageId+" nRecs="+records.size());

	for (int i=0; i<numRecs(); i++) {
	    records.getKey(i).write(out);
	    records.getData(i).write(out);
	}
	// System.out.println("Wrote numRecs="+n);
	byte[] result = byte_out.toByteArray();
	System.arraycopy(result, 0, page, 0, result.length );
	out.close();
	byte_out.close();
    }

    /**
     * @param debugKeys - include key in debug output?
     */
    public void debug (boolean debugKeys) {
	System.out.println("pageId="+pageId);
	System.out.println("pageType="+nodeType);
	System.out.println("level="+level);
	System.out.println("next="+next);
	System.out.println("prev="+prev);
	System.out.println("leastPage="+leastPage);
	System.out.println("availableBytes="+availableBytes);
	System.out.println("numRecs="+numRecs());

	if (debugKeys) {
	    for (int i=0; i<numRecs(); i++) {
		System.out.println(records.getKey(i).toString());
	    }
	}
	System.out.println("");
    }

    /**
     * get a read or write lock on this page.
     * possibly wait
     * @param isWrite - is it a write lock?
     */
    public synchronized void getLock (boolean isWrite) {
	// guard condition
	while (writeLocked || (readLocked && isWrite)) {
	    try {
		wait();
	    } catch (InterruptedException e) {
	        Thread.currentThread().interrupt();
	    }
	}
	if (isWrite) {
	    writeLocked = true;
	} else {
	    nReaders++;
	    readLocked = true;
	}
    }

    /**
     * release the current lock on this page
     */
    public synchronized void releaseLock () {
	if (writeLocked) {
	    writeLocked = false;
	} else {
	    nReaders--;
	    if (nReaders==0) readLocked = false;
	}
	notifyAll();
    }
}
