package ntnustore.btree;

import java.util.ArrayList;
import java.util.List;

import neustore.base.DBRuntimeException;
import neustore.base.Data;
import neustore.base.FloatArrayData;
import neustore.base.FloatData;
import neustore.base.Key;

/**
 * Array to hold a sorted list of Records. The array allows duplicate keys. If
 * you want unique keys, use insertUnique.
 * 
 * @author Svein Erik
 */
public class SortedRecordList<K extends Key<?>, D extends Data<?>> {
    private List<K> keys;
    private List<D> datas;
    private List<Float> datas2;
    private final static int NO_KEY = -2;

    public SortedRecordList() {
        keys = new ArrayList<K>();
        datas = new ArrayList<D>();
        datas2 = new ArrayList<Float>();
    }

    public void clear() {
        keys.clear();
    }

    public int size() {
        return keys.size();
    }

    public K getKey(int pos) {
        return keys.get(pos);
    }

    public D getData(int pos) {
        return datas.get(pos);
    }

    /**
     * Find the position of a given key, or NO_KEY if key is not found.
     * 
     * If leq is true and the key is not found, the position of the first key
     * lower than the given key is returned. If no key exists lower or equal,
     * NO_KEY is returned.
     * 
     * @param key key to search for
     * @param leq
     * @return position of the found key or NO_KEY
     */
    private int find(K key, boolean leq) {
        //System.out.println("keys " + keys);
        //System.out.println("datas get 1 " + datas);
        int low = 0;
        int mid = 0;
        int high = size() - 1;
        // binary search
        while (low <= high) {
            mid = (low + high) / 2;
            int cmp = key.compareTo(keys.get(mid));
            //System.out.println("key " + key);
            //System.out.println("mid " + mid);
            //System.out.println("cmp " + cmp);
            //System.out.println("----------------------------------------------");
            if (cmp == 0) { // key found
            	//System.out.println("2222222222222222222222222222222");
                return mid;
            } else if (cmp < 0) {
                high = mid - 1;
            } else {
                low = mid + 1;
            }
        }

        // not found. For leq either return the first value before key (if one
        // exists) or NO_KEY
        if (leq && (high >= 0)) {
        	//System.out.println("111111111111111111111111");
            return high;
        } else {
            return NO_KEY;
        }
    }

    /**
     * Inserts a key. Allows non-unique keys.
     * 
     * @throws DBRuntimeException if record list is full
     */
    public void insert(K key, D data) {
        int pos = find(key, true);
        pos = (pos == NO_KEY ? 0 : pos + 1);
        //System.out.println(key);
        keys.add(pos, key);
        datas.add(pos, data);
    }

    /**
     * inserts sequentially at the next position and without without check of keys
     * 
     * @throws DBRuntimeException if record list is full
     */
    public void insertSequential(K key, D data) {
        keys.add(key);
        datas.add(data);
    }

    public D find(K key) {
        int pos = find(key, false);
        //System.out.println("datas 1 " +datas);
        if (pos != NO_KEY) {
        	//System.out.println("datas ----------" +datas);
            return datas.get(pos);
        } else {
            return null;
        }
    }
  
    public D findLeq(K key) {
        int pos = find(key, true);
        //System.out.println("pos " +pos);
        //System.out.println("nokey " + NO_KEY);
        if (pos != NO_KEY) {
        	//System.out.println(datas.get(pos));
        	if(datas.get(pos) instanceof FloatArrayData | datas.get(pos) instanceof FloatData){
        	//	//System.out.println("aaaa");
        		return null;
        	}
        	else{
        		return datas.get(pos);
        	}	
        } else {
            return null;
        }
    }
    

    /*
     * Returns position to first record before the search key, so that next() in
     * Cursor will start at first key. May return -1 if we are to start at
     * keys[0].
     */
    public int findForScan(K key) {
        int pos = find(key, true);
        if (pos == NO_KEY) {
            pos = -1;
        }
        // pos is now either on key or on a larger key, scan backwards to find
        // smaller key
        while (pos >= 0) {
            int cmp = key.compareTo(keys.get(pos));
            if (cmp > 0) {
                break;
            }
            pos--;
        }
        return pos;
    }

    /**
     * Removes a record from the given position.
     * 
     * @param pos position of record to remove
     * @return true if record was successfully removed
     * @throws DBRuntimeException if position is our of range (0 <= pos < size())
     */
    public boolean remove(int pos) {
    	//System.out.println("Keys " + keys);
    	//System.out.println("Datas before removeal "  + datas.get(1));
        keys.remove(pos);
        //System.out.println("Keys " + keys);
        datas.remove(pos);
        //System.out.println("Datas get 1"  + datas);
        return true;
    }

    public void removeAllFrom(int pos) {
        keys.get(pos); // forces range checking
        for (int i = keys.size() - 1; i >= pos; i--) {
            remove(i);
        }
    }

    public boolean remove(K key) {
        int pos = find(key, false);
        if (pos == NO_KEY) {
            return false;
        }
        //System.out.println("pos " + pos);
        return remove(pos);
    }
}
