package ntnustore.dlaesa;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import neustore.base.DBRuntimeException;
import neustore.base.Data;
import neustore.base.IntData;
import neustore.base.IntKey;
import neustore.base.Key;
import ntnustore.btree.BTree;
import ntnustore.distance.Distance;

/**
 * TODO: Description of type.
 * 
 * @version $Id:$
 * @author Erik Bagge Ottesen <erikbagg@idi.ntnu.no>
 */
public class KnnCursor<K extends Key<?>, D extends Data<?>> implements Cursor<K, Float> {
    private Cursor<K, Float> cursor;
    private BTree<IntKey, D> dataFile;
    private Distance<D> distance;

    private float rangeLimit;
    private int knnLimit;

    private List<KeyValue<K, Float>> dataQueue;
    private int currentIndex;

    public KnnCursor(Cursor<K, Float> cursor, BTree<IntKey, D> dataFile, Distance<D> distance, float rangeLimit,
            int knnLimit) {
        super();
        this.cursor = cursor;
        this.dataFile = dataFile;
        this.distance = distance;
        this.knnLimit = (knnLimit > 0 ? knnLimit : Integer.MAX_VALUE);
        this.rangeLimit = (rangeLimit > 0 ? rangeLimit : Float.POSITIVE_INFINITY);

    }

    public boolean next() {
        if (dataQueue == null) {
            fetchData();
            currentIndex = 0;
        } else {
            currentIndex++;
        }

        return dataQueue.size() > currentIndex;
    }

    private void fetchData() {
        // fetch and sort entire dataset on increasing lower bound
        List<KeyValue<K, Float>> lowQueue = new ArrayList<KeyValue<K, Float>>();
        while (cursor.next()) {
            lowQueue.add(new KeyValue<K, Float>(cursor.getKey(), cursor.getData()));
        }
        Collections.sort(lowQueue);

        // if no distance calculation is given, we'll give inexact results based
        // on lower bounds
        if (distance == null) {
            dataQueue = lowQueue.subList(0, (lowQueue.size() < knnLimit ? lowQueue.size() : knnLimit));
            return;
        }

        // fill dataQueue with data from lowQueue
        this.dataQueue = new ArrayList<KeyValue<K, Float>>();
        float lowestUpperBound = rangeLimit;

        for (KeyValue<K, Float> keyDist : lowQueue) {
            if ((keyDist.value > rangeLimit) || (keyDist.value >= lowestUpperBound)) {
                // if the lower bound > range limit, we're done.
                // lower bound > Kth lowest real distance, we're done.
                break;
            }

            // calculate real distance
            D data;
            try {
                data = dataFile.search(new IntKey(((IntData) keyDist.key).value()));
            } catch (IOException e) {
                throw new DBRuntimeException(e);
            }
            keyDist.value = distance.distance(data);

            if (keyDist.value <= lowestUpperBound) {
                // distance inside limit
                dataQueue.add(keyDist);

                if (dataQueue.size() > knnLimit) {
                    // dataSet over K elements, remove last entry and set new
                    // bound
                    Collections.sort(dataQueue);
                    dataQueue.remove(knnLimit);
                    lowestUpperBound = dataQueue.get(knnLimit - 1).value;
                }
            }
        }
        Collections.sort(dataQueue);
    }

    public K getKey() {
        return dataQueue.get(currentIndex).key;
    }

    public Float getData() {
        return dataQueue.get(currentIndex).value;
    }

    public static class KeyValue<K, D extends Comparable<D>> extends Pair<K, D> implements Comparable<KeyValue<K, D>> {
        public KeyValue(K key, D value) {
            super(key, value);
        }

        public int compareTo(KeyValue<K, D> o) {
            return value.compareTo(o.value);
        }
    }
}
