package ntnustore.dlaesa;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import neustore.base.DBBuffer;
import neustore.base.DBRuntimeException;
import neustore.base.Data;
import neustore.base.FloatData;
import neustore.base.IntData;
import neustore.base.IntKey;
import ntnustore.btree.BTree;
import ntnustore.distance.Distance;

/**
 * TODO: Description of type.
 * 
 * @version $Id: LaesaBTree.java 66 2010-02-15 13:02:54Z sveinbra $
 * @author Erik Bagge Ottesen <erikbagg@idi.ntnu.no>
 * @author Svein Erik Bratsberg <sveinbra@idi.ntnu.no>
 */
public class LaesaBTree<D extends Data<?>> extends BTree implements LaesaIndex<D> {
    private Distance<D> distance;
    
    private static int numBins = 30;
    private ArrayList<FloatData> statKeys;

    @SuppressWarnings("unchecked")
	public LaesaBTree(DBBuffer buffer, String fileName, Distance<D> distance) throws IOException {
        super(buffer, fileName, !new File(fileName).exists(), new IntData(0), new FloatData(0.0F));

        this.distance = distance;
    }

    @SuppressWarnings("unchecked")
	public void insert(IntKey key, D data) {
        float dist = distance.distance(data);
        //System.out.println("dist="+dist);
        try {
            insert(new FloatData(dist), new IntData(key.value()));
        } catch (IOException e) {
            // TODO shouldn't throw IOException
            throw new DBRuntimeException(e);
        }
    }
    
    @SuppressWarnings("unchecked")
	public float insert2(IntKey key, D data) {
        float dist = distance.distance(data);
        //System.out.println("dist="+dist);
        try {
			super.insert(new IntData(key.value()), new FloatData(dist));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

        return dist;
    }
    
    public Cursor<IntData, Float> query(D query, float range) {
        float dist = distance.distance(query);
        
        Cursor<FloatData, IntData> cursor = query(new FloatData(dist - range), new FloatData(dist + range));

        return new BTreeLowerBoundCursor<IntData>(cursor, dist);
    }
    
    
    /**
     * Private, static inner class used by query() for transforming the Cursor<FloatKey,
     * K> returned from RTree into a Cursor<K, Float> for use in further
     * LAESA-handling.
     * 
     * The Float returned is the greatest lower bound the pivot can calculate
     * for each key.
     */
    private static class BTreeLowerBoundCursor<K> implements Cursor<K, Float> {
        private Cursor<FloatData, K> cursor;
        private float dist;

        public BTreeLowerBoundCursor(Cursor<FloatData, K> cursor, float dist) {
            super();
            this.cursor = cursor;
            this.dist = dist;
        }

        public boolean next() {
            return cursor.next();
        }

        public K getKey() {
            return cursor.getData();
        }

        public Float getData() {
            FloatData data = cursor.getKey();
            return Math.abs(data.value() - dist);
        }

    }

    public float costOfQuery (D query, float range) {
        float dist = distance.distance(query);
        return costOfScan(new FloatData(dist - range), new FloatData(dist + range));
    }
    
    /**
     * Estimate the cost of the scan.
     * @param fromKey
     * @param toKey
     * @return estimated fraction of tree to scan
     */
    public float costOfScan (FloatData fromKey, FloatData toKey) {
        int lo = 0;
        int hi = numBins;
        for (int i=0;i<numBins;i++) {
            if (statKeys.get(i).compareTo(fromKey)<0) {
                lo = i;
            }
            if (hi==numBins && toKey.compareTo(statKeys.get(i))<0) {
                hi = i;
            }
        }
        return (float)(hi-lo)/numBins;
    }

    /**
     * Generate statistics for query cost estimation
     */
    public void generateStatistics () {
 
        try {
            statKeys = createEquivDepthStatistics(numBins);
        } catch (IOException e) {
            // TODO shouldn't throw IOException
            throw new DBRuntimeException(e);
        }
    }

}
