package ntnustore.dlaesa;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Comparator;
import java.util.Arrays;
import java.util.concurrent.ExecutorService;

import neustore.base.Data;
import neustore.base.FloatData;
import neustore.base.FloatArrayData;
import neustore.base.IntData;
import neustore.base.IntKey;
import ntnustore.btree.BTree;
import ntnustore.distance.Distance;
import ntnustore.distance.L2Distance;


/**
 * TODO: Description of type.
 * 
 * @version $Id:$
 * @author Erik Bagge Ottesen <erikbagg@idi.ntnu.no>
 * @author Svein Erik Bratsberg <sveinbra@idi.ntnu.no>
 */
public class LaesaIndexList<D extends Data<?>> implements LaesaIndex<D> {
    private List<LaesaIndex<D>> indexes;

    public LaesaIndexList() {
        super();
        this.indexes = new ArrayList<LaesaIndex<D>>();
    }

    public boolean add(LaesaIndex<D> e) {
        return this.indexes.add(e);
    }

    public int size() {
        return this.indexes.size();
    }

    public void insert(IntKey key, D data) {
        for (LaesaIndex<D> l : indexes) {
            l.insert(key, data);
        }
    }
    
    public Cursor<IntData, Float> queryBest (D query, float range) {
        float bestCost = Float.POSITIVE_INFINITY;
        int bestIdx = -1;
        for (int i=0; i<size(); i++) {
            float cost = indexes.get(i).costOfQuery(query, range); 
            //System.out.println("idx="+i+" cost="+cost);
            if (cost<bestCost) {
                bestIdx = i;
                bestCost = cost;
            }
        }
        return indexes.get(bestIdx).query(query, range);
    }
    public Cursor<IntData, Float> query2Best (D query, float range) {
        int best1idx = -1;
        int best2idx = -1;
        float best1cost = Float.POSITIVE_INFINITY;
        float best2cost = Float.POSITIVE_INFINITY;
        for (int i=0; i<size(); i++) {
           float cost = indexes.get(i).costOfQuery(query, range);
           //System.out.println("idx=" + i + " cost="+cost);
           if (cost<best1cost) {
               best2idx = best1idx;
               best2cost = best1cost;
               best1idx = i;
               best1cost = cost;
           } else if (cost<best2cost) {
               best2idx = i;
               best2cost = cost;
           }
        }
        //System.out.println("best1=" + best1idx + " best2=" + best2idx);
        Cursor<IntData, Float> cursor = indexes.get(best1idx).query(query, range);
        Cursor<IntData, Float> outerCursor = indexes.get(best2idx).query(query, range);
        return new MaxFloatHashJoin<IntData>(cursor, outerCursor);
    }

    private static class IdxCost {
        public int idx;
        public float cost;
    }
    private class IdxCostComparator implements Comparator<IdxCost> {
        public int compare (IdxCost i1, IdxCost i2) {
            if (i1.cost<i2.cost) {
                return -1;
            } else if (i1.cost==i2.cost) {
                return 0;
            } else {
                return 1;
            }
        }
    }
    public Cursor<IntData, Float> queryKBest (D query, float range, int k) {
        IdxCost[] idxCost = new IdxCost[size()];
        for (int i=0; i<size(); i++) {
            idxCost[i] = new IdxCost();
            idxCost[i].idx = i;
            idxCost[i].cost = indexes.get(i).costOfQuery(query, range);
        }
        Arrays.sort(idxCost, new IdxCostComparator());
        //for (int i=0; i<size(); i++) {
        //    System.out.println("idx="+idxCost[i].idx+" cost="+idxCost[i].cost);
        //}
        Cursor<IntData, Float> cursor = indexes.get(idxCost[0].idx).query(query, range);
        for (int i = 1; i < k; i++) {
            Cursor<IntData, Float> outerCursor = indexes.get(idxCost[i].idx).query(query, range);
            cursor = new MaxFloatHashJoin<IntData>(cursor, outerCursor);

        }
        return cursor;
    }

    public void queryCountEachIdx(D query, float range) {
        for (int i=0; i<size(); i++) {
            Cursor<IntData, Float> cursor = indexes.get(i).query(query, range);
     	    int count = 0;
            // count results
            while (cursor.next()) {
                count++;
            }
            System.out.println("pivot="+i+" count="+count+" address="+indexes.get(i));
        }
    }

    public Cursor<IntData, Float> query(D query, float range) {
        Cursor<IntData, Float> cursor = indexes.get(0).query(query, range);

        for (int i = 1; i < indexes.size(); i++) {
            Cursor<IntData, Float> outerCursor = indexes.get(i).query(query, range);
            cursor = new MaxFloatHashJoin<IntData>(cursor, outerCursor);
        }

        return cursor;
    }

    public Cursor<IntData, Float> query(D query, float range, ExecutorService executor, int threadCount) {
        // create <threadCount> left-deep join-trees
        List<Cursor<IntData, Float>> cursors = new ArrayList<Cursor<IntData, Float>>();

        threadCount = Math.min(threadCount, indexes.size());
        int currentPivot = 0;
        for (int t = 0; t < threadCount; t++) {
            Cursor<IntData, Float> cursor = indexes.get(currentPivot).query(query, range);
            currentPivot++;

            int count = indexes.size() / threadCount + (t < indexes.size() % threadCount ? 1 : 0);
            for (int i = 1; i < count; i++) {
                Cursor<IntData, Float> outerCursor = indexes.get(currentPivot).query(query, range);
                currentPivot++;
                cursor = new MaxFloatHashJoin<IntData>(cursor, outerCursor);
            }

            BufferedCursor<IntData, Float> bufferedCursor = new BufferedCursor<IntData, Float>(cursor,
                    Integer.MAX_VALUE);
            executor.execute(bufferedCursor);
            cursors.add(bufferedCursor);
        }

        // left-deep join of the join-trees
        Cursor<IntData, Float> cursor = cursors.remove(0);
        while (cursors.size() > 0) {
            Cursor<IntData, Float> outerCursor = cursors.remove(0);
            cursor = new MaxFloatHashJoin<IntData>(cursor, outerCursor);
        }

        // BufferedCursor<IntData, Float> bufferedCursor = new
        // BufferedCursor<IntData, Float>(cursor, Integer.MAX_VALUE);
        // executor.execute(bufferedCursor);

        return cursor;
    }

    public void close() throws IOException {
        for (LaesaIndex<D> l : indexes) {
            l.close();
        }
        indexes.clear();
    }

    /**
     * Find and returns the best pivot candidate according to a modified version
     * of the pivot selection algorithm in "Roberto Figueira Santos Filho, Agma
     * Traina, C. Traina, and C. Faloutsos. Similarity search without tears: The
     * omni family of all-purpose access methods. Proceedings of the 17th
     * International Conference on Data Engineering, pages 623--630, April 2-6
     * 2001. Heidelberg, Germany."
     * 
     * The algorithm is modified to only maximize distance from the current set
     * of pivots to the new candidate, and not from an edge. This simplifies
     * calculation using the existing pivot structure greatly. Note that it is
     * the minimum distance that is maximized, that is the distance to the
     * closest pivot is maximized.
     * 
     * Multiple calls without the addition of new pivots or data objects will
     * simply return the same pivot candidate. As such this method is
     * deterministic when used with the same pivot and data set.
     * 
     * @return a pivot candidate from the current data set
     * @throws IOException
     */
    public IntKey getPivotCandidate() throws IOException {
        LaesaBTree<?> firstIndex = (LaesaBTree<?>) indexes.get(0);

        // scan first index and create a candidate list
        Set<Integer> candidates = new HashSet<Integer>();
        Cursor<FloatData, IntData> keyCursor = firstIndex.query(new FloatData(0.0F), new FloatData(Float.MAX_VALUE));
        while (keyCursor.next()) {
            int key = keyCursor.getData().value();
            candidates.add(key);
        }

        // create cursors for scan of all pivot indexes
        float minDist = Float.NEGATIVE_INFINITY;
        List<Cursor<FloatData, IntData>> cursors = new ArrayList<Cursor<FloatData, IntData>>();

        for (LaesaIndex<D> index : indexes) {
            LaesaBTree<?> bindex = (LaesaBTree<?>) index;
            Cursor<FloatData, IntData> cursor = bindex.query(new FloatData(0.0F), new FloatData(Float.MAX_VALUE));
            if (cursor.next()) {
                minDist = Math.min(minDist, cursor.getKey().value());
                cursors.add(cursor);
            }
        }

        // scan all remaining cursors simultaneously and remove candidates until
        // only one remains
        Map<Integer, Float> finalCandidates = new HashMap<Integer, Float>();
        float newMinDist = minDist;

        while (candidates.size() > 0) {
            finalCandidates.clear();

            for (int i = cursors.size() - 1; i >= 0; i--) {
                Cursor<FloatData, IntData> cursor = cursors.get(i);

                // scan cursor until we hit a distance > minDist
                float dist = cursor.getKey().value();
                while (dist <= minDist) {
                    // add all keys that are still candidates to
                    // finalCandidates, with the minimum known distance
                    int key = cursor.getData().value();
                    if (candidates.contains(key)) {
                        if (finalCandidates.containsKey(key)) {
                            finalCandidates.put(key, Math.min(dist, finalCandidates.get(key)));
                        } else {
                            finalCandidates.put(key, dist);
                        }
                    }
                    // next tuple
                    if (!cursor.next()) {
                        cursors.remove(i);
                        break;
                    }
                    dist = cursor.getKey().value();
                }

                // update new minimum distance for next round
                newMinDist = Math.max(dist, newMinDist);
            }
            // new minDist
            minDist = newMinDist;

            // remove all candidates found in this round
            candidates.removeAll(finalCandidates.keySet());
        }

        int maxKey = -1;
        float maxDist = Float.NEGATIVE_INFINITY;
        for (int key : finalCandidates.keySet()) {
            Float dist = finalCandidates.get(key);
            if (dist > maxDist) {
                maxDist = dist;
                maxKey = key;
            }
        }

        return new IntKey(maxKey);
    }

    /**
     * Find and returns the best pivot candidate according to a modified version
     * of the pivot selection algorithm in "Roberto Figueira Santos Filho, Agma
     * Traina, C. Traina, and C. Faloutsos. Similarity search without tears: The
     * omni family of all-purpose access methods. Proceedings of the 17th
     * International Conference on Data Engineering, pages 623--630, April 2-6
     * 2001. Heidelberg, Germany."
     * 
     * The algorithm is modified to only maximize distance from the current set
     * of pivots to the new candidate, and not from an edge. This simplifies
     * calculation using the existing pivot structure greatly. Note that it is
     * the minimum distance that is maximized, that is the distance to the
     * closest pivot is maximized.
     * 
     * This returns a list of pivots from the given dataFile. This works much
     * faster if all of dataFile can fit in main memory.
     * 
     * @return a pivot candidate from the current data set
     */
    public List<IntKey> getPivotCandidate(BTree<IntKey, D> dataFile, Distance<D> distance, int pivotCount) throws IOException {
        List<IntKey> pivots = new ArrayList<IntKey>();
        Map<IntKey, Float> minDist = new HashMap<IntKey, Float>();

        while (pivots.size() < pivotCount) {
            float maxDist = 0.0F;
            IntKey maxKey = null;
            D maxData = null;

            // Scan entire datafile
            Cursor<IntKey, D> cursor = dataFile.query(new IntKey(0), new IntKey(Integer.MAX_VALUE));
            while (cursor.next()) {
                // calculate the distance from the new pivot to each data point
            	System.out.println(cursor.getKey());
                IntKey key = cursor.getKey();
                float dist = distance.distance(cursor.getData());
                if (distance.getCalculations() % 1000 == 0) {
                	//System.out.println(distance.getCalculations());
                }

                // update minimum distance
                if (pivots.size() == 0) {
                    minDist.put(key, Float.POSITIVE_INFINITY);
                } else {
                    dist = Math.min(dist, minDist.get(key));
                    minDist.put(key, dist);
                }

                // find farthest data point
                if (dist > maxDist) {
                    maxDist = dist;
                    maxKey = key;
                    maxData = cursor.getData();
                }
            }
            
            //System.out.println(maxKey);
            // greatest minimum distance = new pivot
            pivots.add(maxKey);
            distance.initialize(maxData);
        }

        return pivots;
    }

    /**
     * Finds and returns the objects farthest away from the given distance. This
     * is based on the algorithm for finding the first pivot as described in
     * "Roberto Figueira Santos Filho, Agma Traina, C. Traina, and C. Faloutsos.
     * Similarity search without tears: The omni family of all-purpose access
     * methods. Proceedings of the 17th International Conference on Data
     * Engineering, pages 623--630, April 2-6 2001. Heidelberg, Germany."
     * 
     * Multiple calls without the addition of new pivots or data objects will
     * simply return the same pivot candidate. As such this method is
     * deterministic when used with the same pivot and data set.
     * 
     * @return
     * 
     * @return a pivot candidate from the current data set
     * @throws IOException
     */
    public IntKey getPivotCandidate(Cursor<IntKey, D> cursor, Distance<D> distance) throws IOException {
        float maxDist = 0.0F;
        IntKey maxKey = null;

        // Scan entire datafile
        while (cursor.next()) {
            // calculate the distance from the new pivot to each datapoint
            IntKey key = cursor.getKey();
            float dist = distance.distance(cursor.getData());

            if (dist > maxDist) {
                maxDist = dist;
                maxKey = key;
            }
        }

        return maxKey;
    }

    public float costOfQuery(D query, float range) {
        float cost = 0.0F;
        for (int i = 0; i < size(); i++) {
            cost += indexes.get(i).costOfQuery(query, range);
        }
        return cost;
    }
    /**
     * Generate statistics for each index.
     */
    public void generateStatistics () {
        for (int i = 0; i < size(); i++) {
            indexes.get(i).generateStatistics();
        }
    }

	@Override
	public float insert2(IntKey key, D data) {
		// TODO Auto-generated method stub
		return 0;
	}
}
