package ntnustore.dlaesa;

import java.io.File;
import java.io.IOException;
import java.util.List;

import neustore.base.DBBuffer;
import neustore.base.IntData;
import neustore.base.IntKey;
import ntnustore.rtree.RTreeRecord;
import ntnustore.rtree.RTreeRegion;
import ntnustore.distance.Distance;
import ntnustore.rtree.RTree;

/**
 * LAESA methods based on distances from several pivots stored in a single
 * RTree. Support for insert and range query.
 * 
 * @version $Id:$
 * @author Erik Bagge Ottesen <erikbagg@idi.ntnu.no>
 */
public class LaesaRTree<D> extends RTree implements LaesaIndex<D> {
    private List<Distance<D>> distances;

    public LaesaRTree(DBBuffer buffer, String fileName, List<Distance<D>> distances) throws IOException {
        super(buffer, fileName, !new File(fileName).exists(), distances.size(), new IntData(0));

        this.distances = distances;
    }

    public void insert(IntKey key, D data) {
        float[] dists = getDistances(data);

        RTreeRecord<IntData> record = new RTreeRecord<IntData>(new RTreeRegion(dists), new IntData(key.value()));
        insert(record);
    }

    private float[] getDistances(D data) {
        float[] dists = new float[distances.size()];
        for (int i = 0; i < dists.length; i++) {
            dists[i] = distances.get(i).distance(data);
        }
        return dists;
    }

    public Cursor<IntData, Float> query(D query, float range) {
        float[] dists = getDistances(query);

        // query query object for RTree.query() and execute query
        RTreeRegion queryRange = new RTreeRegion(dists.length);
        for (int i = 0; i < dists.length; i++) {
            queryRange.low[i] = dists[i] - range;
            queryRange.high[i] = dists[i] + range;
        }

        Cursor<RTreeRegion, IntData> cursor = query(queryRange);
        Cursor<IntData, Float> lowerBoundCursor = new RTreeLowerBoundCursor<IntData>(cursor, dists);

        return lowerBoundCursor;
    }

    /**
     * Private, static inner class used by query() for transforming the Cursor<RTreeRegion,
     * K> returned from RTree into a Cursor<K, LowHighData> for use in further
     * LAESA-handling.
     * 
     * This returned the greatest lower bound as well as the lowest upper bound
     * of all the dimensions of the RTree.
     */
    private static class RTreeLowerBoundCursor<K> implements Cursor<K, Float> {
        private Cursor<RTreeRegion, K> cursor;
        private float[] dists;

        public RTreeLowerBoundCursor(Cursor<RTreeRegion, K> cursor, float[] dists) {
            super();
            this.cursor = cursor;
            this.dists = dists;
        }

        public boolean next() {
            return cursor.next();
        }

        public K getKey() {
            return cursor.getData();
        }

        public Float getData() {
            RTreeRegion data = cursor.getKey();

            float low = Float.NEGATIVE_INFINITY;
            for (int i = 0; i < data.dimensions; i++) {
                float dist = Math.abs(data.low[i] - dists[i]);
                low = Math.max(low, dist);
            }

            return low;
        }
    }

    /**
     * Return the estimated cost of the range query
     * @param query
     * @param range
     * @return
     */
    public float costOfQuery (D query, float range) {
        float[] dists = getDistances(query);

        RTreeRegion queryRange = new RTreeRegion(dists.length);
        for (int i = 0; i < dists.length; i++) {
            queryRange.low[i] = dists[i] - range;
            queryRange.high[i] = dists[i] + range;
        }
        return costOfQuery(queryRange);
    }

    /**
     * TODO: Generate statistics for query cost estimation
     */
    public void generateStatistics () {
    }

	@Override
	public float insert2(IntKey key, D data) {
		// TODO Auto-generated method stub
		return 0;
	}


}
