package ntnustore.rtree;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayDeque;
import java.util.Deque;

import neustore.base.DBBuffer;
import neustore.base.DBBufferReturnElement;
import neustore.base.DBIndex;
import neustore.base.DBRuntimeException;
import neustore.base.IntData;
import ntnustore.dlaesa.Cursor;

/**
 * @version $Id: RTree.java 72 2010-03-11 13:29:28Z sveinbra $
 * @author Erik Bagge Ottesen <erikbagg@idi.ntnu.no>
 */
public class RTree extends DBIndex {
    private int rootNode;
    private int numRecs;
    private int dimensions;

    private IntData sampleData;

    public RTree(DBBuffer buffer, String filename, boolean create, int dimensions, IntData sampleData)
            throws IOException {
        super(buffer, filename, create);

        this.sampleData = sampleData;
        this.dimensions = dimensions;

        if (create) {
            rootNode = allocate();
            RTreePage rootPage = new RTreePage(RTreePage.RTREE_PAGE, rootNode, 0, pageSize, dimensions, sampleData);
            writePage(rootPage);
        }
    }

    /**
     * Chooses the optimal subtree to include the given mbr at the given level.
     * 
     * @param mbr the mbr of the record to be inserted
     * @param level the level to insert the record
     * @param path will be filled with the path of all parents above node in the
     *        tree
     * @return the optimal insertion node
     */
    private RTreePage chooseSubtree(RTreeRegion mbr, int level, Deque<RTreePage> path) {
        RTreePage node = readPage(rootNode);

        while (node.level > level) {
            path.push(node);
            int childId = node.chooseSubtree(mbr);
            node = readPage(childId);
        }

        return node;
    }

    /**
     * Insert a record into the given node with the given parent path.
     * 
     * @param node insert record into this node
     * @param record to be inserted
     * @param path the path of parents over the child
     */
    private void insert(RTreePage node, RTreeRecord<IntData> record, Deque<RTreePage> path) {
        if (!node.isFull()) {
            // node is not full, insert
            RTreeRegion nodeMbr = node.getMinimumBoundingRectangle();

            node.insert(record);
            writePage(node);

            // adjust covering rectangles of parents if the record mbr is not
            // within the old node mbr
            if (!nodeMbr.contains(record.mbr)) {
                adjustMbrs(node, path);
            }
        } else {

            // node is full, allocate a new page at this level
            int newPageId = allocate();
            RTreePage newPage = new RTreePage(RTreePage.RTREE_PAGE, newPageId, node.level, pageSize, dimensions,
                    sampleData);

            // insert the record and then split the page (ensures the new record
            // is in the correct node after split)
            node.insertSplit(record, newPage);

            // adjust covering rectangles of parents
            adjustMbrs(node, path);

            // if we just split the root node we need a new root
            if (node.pageId == rootNode) {
                // create a new root
                rootNode = allocate();
                RTreePage rootPage = new RTreePage(RTreePage.RTREE_PAGE, rootNode, node.level + 1, pageSize,
                        dimensions, sampleData);
                writePage(rootPage);
                // insert old root into new root
                insertPage(node);
                //System.out.println("inserting root node at level="+(node.level+1));
            }

            // store and insert split page
            writePage(node);
            writePage(newPage);
            insertPage(newPage);
        }
    }
    
    /**
     * Estitame the cost of query
     * @param mbr
     * @return
     */
    public float costOfQuery (RTreeRegion mbr) {
        RTreePage root = readPage(rootNode);
        return (float)root.findFractionOfOverlaps(mbr);
    }

    /**
     * Adjusts the bounding boxes of all parents in the given path so that they
     * are minimally covering.
     * 
     * @param node the node that might have changed mbr
     * @param path the path of parents over the child
     */
    private void adjustMbrs(RTreePage node, Deque<RTreePage> path) {
        for (RTreePage parent : path) {
            boolean changed = parent.updateMbr(node.pageId, node.getMinimumBoundingRectangle());
            if (!changed) {
                break;
            }
            writePage(parent);

            node = parent;
        }
    }

    private void insert(RTreeRecord<IntData> record, int level) {
        Deque<RTreePage> path = new ArrayDeque<RTreePage>();
        RTreePage node = chooseSubtree(record.mbr, level, path);

        insert(node, record, path);
        numRecs++;
    }

    /**
     * Inserts a record into the RTree
     * 
     * @param record the record to be inserted
     */
    public void insert(RTreeRecord<IntData> record) {
        insert(record, 0); // data is at level 0
    }

    /**
     * Inserts a page into the current RTree with the pageid, mbr and level of
     * the actual page.
     * 
     * @param page the page to be inserted
     */
    private void insertPage(RTreePage page) {
        insert(new RTreeRecord<IntData>(page.getMinimumBoundingRectangle(), new IntData(page.pageId)), page.level + 1);
    }

    /**
     * Read the given page from buffer or disk.
     * 
     * @param pageId of page to read
     * @return RTree-page with the given pageId.
     */
    protected RTreePage readPage(int pageId) {
        try {
            DBBufferReturnElement page = buffer.readPage(file, pageId);
            if (page.parsed) {
                return (RTreePage) page.object;
            } else {
                // read pageType to check whether to create index or leaf page
                byte[] pageData = (byte[]) page.object;
                RTreePage result = new RTreePage(RTreePage.RTREE_PAGE, pageId, -1, pageSize, dimensions, sampleData);
                result.read(pageData);
                return result;
            }
        } catch (IOException e) {
            throw new DBRuntimeException("Error while reading page: " + pageId, e);
        }
    }

    /**
     * Write the current page to buffer. That is mark the current page as dirty
     * so it will be written to disk at a later time.
     * 
     * @param page to write to buffer
     */
    private void writePage(RTreePage page) {
        try {
            buffer.writePage(file, page.pageId, page);
        } catch (IOException e) {
            throw new DBRuntimeException("Error while writing page: " + page, e);
        }
    }

    public Cursor<RTreeRegion, IntData> query(RTreeRegion query) {
        return new RTreeRangeQueryCursor(this, query, rootNode);
    }

    @Override
    protected void initIndexHead() throws IOException {
        // TODO: initIndexHead() is useless as long as it called from the
        // super() constructor.
    }

    @Override
    protected void readIndexHead(byte[] indexHead) throws IOException {
        System.out.println("RTree.readIndexHead()");
        DataInputStream in = new DataInputStream(new ByteArrayInputStream(indexHead));

        numRecs = in.readInt();
        rootNode = in.readInt();
        dimensions = in.readInt();
    }

    @Override
    protected void writeIndexHead(byte[] indexHead) throws IOException {
        ByteArrayOutputStream byte_out = new ByteArrayOutputStream(pageSize);
        DataOutputStream out = new DataOutputStream(byte_out);

        out.writeInt(numRecs);
        out.writeInt(rootNode);
        out.writeInt(dimensions);

        out.close();
        byte[] result = byte_out.toByteArray();
        System.arraycopy(result, 0, indexHead, 0, result.length);
    }

}
