package ntnustore.rtree;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Collections;

import neustore.base.DBPage;
import neustore.base.DBRuntimeException;
import neustore.base.IntData;

/**
 * TODO: Description of type.
 * 
 * @version $Id: RTreePage.java 68 2010-02-19 08:56:14Z sveinbra $
 * @author Erik Bagge Ottesen <erikbagg@idi.ntnu.no>
 */
public class RTreePage extends DBPage {
    private static final boolean LEAF_RECORDS_POINT_KEYS = true;

    protected final int RESERVED = 36;
    protected static final int RTREE_PAGE = 12;

    protected int pageId;
    protected int level; // 0=leaf
    protected int availableBytes;

    // lock variables
    protected boolean writeLocked;
    protected boolean readLocked;
    protected int nReaders;

    // int numRecs; is stored on disk, but not in memory

    protected int dimensions;
    protected int keySize;

    private RTreeRecordList<IntData> records;

    protected IntData sampleData;

    public RTreePage(int pageType, int pageId, int level, int pageSize, int dimensions, IntData sampleData) {
        super(pageType, pageSize);
        this.pageId = pageId;
        this.level = level;
        this.records = new RTreeRecordList<IntData>();
        this.dimensions = dimensions;
        this.sampleData = sampleData;

        this.keySize = (LEAF_RECORDS_POINT_KEYS && (level == 0) ? 4 : 8) * dimensions;
        this.availableBytes = pageSize - RESERVED;
    }

    public int size() {
        return records.size();
    }

    public boolean isFull() {
        // return size() > 10;
        return (availableBytes < keySize + sampleData.maxSize());
    }

    public void split(RTreePage newPage) {
        int M = size();
        int m = M * 4 / 10; // "m = 40% yields the best performance"
        if (m < 2) {
            throw new DBRuntimeException("Invalid split: M=" + M + ", m=" + m);
        }

        int axis = chooseSplitAxis(M, m);

        RTreeRecordList<IntData> resultGroup1 = new RTreeRecordList<IntData>();
        RTreeRecordList<IntData> resultGroup2 = new RTreeRecordList<IntData>();
        chooseSplitIndex(M, m, axis, resultGroup1, resultGroup2);

        this.clear();
        this.insertAll(resultGroup1);
        newPage.clear();
        newPage.insertAll(resultGroup2);
    }

    public void insertSplit(RTreeRecord<IntData> record, RTreePage newPage) {
        records.add(record);
        split(newPage);
    }

    private int chooseSplitAxis(int M, int m) {
        double minMargin = Double.POSITIVE_INFINITY;
        int minAxis = 0;

        RTreeRecordList<IntData> records = new RTreeRecordList<IntData>(this.records);

        for (int axis = 0; axis < dimensions; axis++) {
            for (int low = 0; low < 2; low++) {
                Collections.sort(records, new RTreeRecordComparator(axis, low == 1));

                for (int k = m; k <= M - m; k++) {
                    RTreeRegion u1 = records.getMinimumBoundingRectangle(0, k);
                    RTreeRegion u2 = records.getMinimumBoundingRectangle(k, M);

                    double margin = u1.getMargin() + u2.getMargin();

                    if (margin < minMargin) {
                        minMargin = margin;
                        minAxis = axis;
                    }
                }
            }
        }

        return minAxis;
    }

    private void chooseSplitIndex(int M, int m, int axis, RTreeRecordList<IntData> resultGroup1,
            RTreeRecordList<IntData> resultGroup2) {
        double minOverlap = Double.POSITIVE_INFINITY;
        double minArea = Double.POSITIVE_INFINITY;

        RTreeRecordList<IntData> records = new RTreeRecordList<IntData>(this.records);

        for (int low = 0; low < 2; low++) {
            Collections.sort(records, new RTreeRecordComparator(axis, low == 1));

            for (int k = m; k <= M - m; k++) {
                RTreeRegion u1 = records.getMinimumBoundingRectangle(0, k);
                RTreeRegion u2 = records.getMinimumBoundingRectangle(k, M);

                RTreeRegion intersection = RTreeRegion.intersect(u1, u2);
                double overlap = intersection.getArea();
                double area = u1.getArea() + u2.getArea();

                if ((overlap < minOverlap) || ((overlap == minOverlap) && (area < minArea))) {
                    minOverlap = overlap;
                    minArea = area;

                    resultGroup1.clear();
                    resultGroup1.addAll(records.subList(0, k));
                    resultGroup2.clear();
                    resultGroup2.addAll(records.subList(k, M));
                }
            }
        }

    }

    private void insertAll(RTreeRecordList<IntData> records) {
        for (RTreeRecord<IntData> record : records) {
            insert(record);
        }
    }

    public void clear() {
        records.clear();
        availableBytes = pageSize - RESERVED;
    }

    public void insert(RTreeRecord<IntData> record) {
        if (record.mbr.dimensions != dimensions) {
            throw new DBRuntimeException("mbr.dimensions != dimensions: " + record.mbr.dimensions + " - " + dimensions);
        }

        availableBytes -= keySize + record.data.size();
        if (availableBytes < 0) {
            throw new DBRuntimeException();
        }

        records.add(record);
    }

    public RTreeRegion getMinimumBoundingRectangle() {
        if (size() == 0) {
            return new RTreeRegion(dimensions);
        } else {
            return records.getMinimumBoundingRectangle();
        }
    }

    //
    // Index methods
    //

    private RTreeRecord<IntData> getRecord(int objectId) {
        for (RTreeRecord<IntData> r : records) {
            if (r.data.value() == objectId) {
                return r;
            }
        }
        return null;
    }

    public boolean updateMbr(int pageId, RTreeRegion mbr) {
        RTreeRecord<IntData> child = getRecord(pageId);
        if (child == null) {
            throw new DBRuntimeException("Child not in page");
        }

        if (!child.mbr.equals(mbr)) {
            child.mbr = mbr;
            return true;
        } else {
            return false;
        }
    }

    public int chooseSubtree(RTreeRegion mbr) {
        if (level < 1) {
            throw new DBRuntimeException("chooseSubtree at level " + level);
        }
        if (level == 1) {
            // if the child pointers in N point to leaves [determine the minimum
            // overlap cost]
            return findMinimumOverlap(mbr);
        } else {
            // if the child pointers in N do not point to leaves [determine the
            // minimum area cost]
            return findMinimumArea(mbr);
        }
    }

    public double findFractionOfOverlaps (RTreeRegion mbr) {
       double overlapArea = 0.0F;
       for (RTreeRecord<IntData> child : records) {
           if (RTreeRegion.intersects(child.mbr, mbr)) {
               double interArea = RTreeRegion.intersect(child.mbr, mbr).getArea();
               //System.out.println("interArea=" + interArea + " child.area=" + child.mbr.getArea());
               overlapArea += interArea / child.mbr.getArea();
           }
       }
       return overlapArea;
    }
    
    public int findMinimumOverlap(RTreeRegion mbr) {
        // TODO: actually implement method
        // Note: We have chosen not to implement minimum-overlap as Beckmann
        // states this has a minimal effect on number of file reads and has a
        // higher cpu-cost. As our experiments for R-trees are clearly
        // cpu-bound, this will likely lead to slower, and not faster, response
        // times both for indexing and queries.
        return findMinimumArea(mbr);
    }

    protected int findMinimumArea(RTreeRegion mbr) {
        int minChildPageId = -1;
        double minUnionArea = Double.POSITIVE_INFINITY;
        double minEnlargementArea = Double.POSITIVE_INFINITY;

        for (RTreeRecord<IntData> child : records) {
            double unionArea = RTreeRegion.union(child.mbr, mbr).getArea();
            double enlargement = unionArea - child.mbr.getArea();

            // choose the entry in N whose rectangle needs least area
            // enlargement to include the new data rectangle.
            // resolve ties by choosing the entry with the rectangle of
            // the smallest area
            if ((enlargement < minEnlargementArea)
                    || ((enlargement == minEnlargementArea) && (unionArea < minUnionArea))) {
                minEnlargementArea = enlargement;
                minChildPageId = child.data.value();
                minUnionArea = child.mbr.getArea();
            }
        }

        return minChildPageId;
    }

    public RTreeRecordList<IntData> getRecords() {
        return this.records;
    }

    @Override
    protected void read(byte[] page) throws IOException {
        ByteArrayInputStream byte_in = new ByteArrayInputStream(page);
        DataInputStream in = new DataInputStream(byte_in);
        nodeType = in.readInt();
        pageSize = in.readInt();
        pageId = in.readInt();
        level = in.readInt();
        availableBytes = in.readInt();
        int numRecs = in.readInt();

        keySize = (LEAF_RECORDS_POINT_KEYS && (level == 0) ? 4 : 8) * dimensions;

        clear();
        for (int i = 0; i < numRecs; i++) {
            RTreeRegion mbr = new RTreeRegion(dimensions);
            for (int d = 0; d < dimensions; d++) {
                mbr.low[d] = in.readFloat();
                if (LEAF_RECORDS_POINT_KEYS && (level == 0)) {
                    mbr.high[d] = mbr.low[d];
                } else {
                    mbr.high[d] = in.readFloat();
                }
            }

            IntData data = (IntData) sampleData.clone();
            data.read(in);

            insert(new RTreeRecord<IntData>(mbr, data));
        }

        in.close();
        byte_in.close();
    }

    @Override
    protected void write(byte[] page) throws IOException {
        ByteArrayOutputStream byte_out = new ByteArrayOutputStream(pageSize);
        DataOutputStream out = new DataOutputStream(byte_out);
        out.writeInt(nodeType);
        out.writeInt(pageSize);
        out.writeInt(pageId);
        out.writeInt(level);
        out.writeInt(availableBytes);
        out.writeInt(records.size());

        for (RTreeRecord<IntData> record : records) {
            for (int i = 0; i < record.mbr.dimensions; i++) {
                out.writeFloat(record.mbr.low[i]);
                if (LEAF_RECORDS_POINT_KEYS && (level == 0)) {
                } else {
                    out.writeFloat(record.mbr.high[i]);
                }
            }
            record.data.write(out);
        }

        out.close();
        byte[] result = byte_out.toByteArray();
        System.arraycopy(result, 0, page, 0, result.length);
    }
}
