package test;

import java.io.FileWriter;
import java.io.IOException;
import java.util.List;

import neustore.base.Data;
import neustore.base.IntKey;
import neustore.base.LRUBuffer;
import ntnustore.btree.BTree;
import ntnustore.distance.Distance;
import ntnustore.dlaesa.Cursor;
import ntnustore.dlaesa.LaesaBTree;
import ntnustore.dlaesa.LaesaHeapFile;
import ntnustore.dlaesa.LaesaIndex;
import ntnustore.dlaesa.LaesaIndexList;
import ntnustore.dlaesa.LaesaRTree;

public class AbstractTest<T extends Data<?>> {
	// INDEX_TYPE: -1=HeapFile, 0=BTree, 1+=RTree (number indicates maximum
	// dimensionality of R-tree)
	private int indexType;

	protected LRUBuffer buffer;
	protected BTree<IntKey, T> dataFile;
	protected LaesaIndexList<T> indexList = new LaesaIndexList<T>();

	private List<Distance<T>> pivotList;
	private int pivotCount;

	private int dataCount;

	private StringBuffer logBuffer;

	public void init(int indexType, int pageSize, int bufferSize, T sampleData,
			List<Distance<T>> pivotList) throws IOException {
		TestUtil.deleteFiles();

		buffer = new LRUBuffer(bufferSize, pageSize);
		dataFile = new BTree<IntKey, T>(buffer, "data/test", true,
				new IntKey(0), sampleData);
		indexList = new LaesaIndexList<T>();
		logBuffer = new StringBuffer();

		this.indexType = indexType;
		this.pivotList = pivotList;
		this.pivotCount = 0;
		this.dataCount = 0;
	}

	public void close() throws IOException {
		indexList.close();
		dataFile.close();
		TestUtil.deleteFiles();
	}

	public void addPivot() throws IOException {
		// create index
		LaesaIndex<T> index;
		if (indexType > 0) {
			List<Distance<T>> distances = pivotList.subList(pivotCount,
					pivotCount + indexType);
			index = new LaesaRTree<T>(buffer, "data/test_pivot" + pivotCount,
					distances);
			pivotCount += indexType;
		} else if (indexType == 0) {
			Distance<T> distance = pivotList.get(pivotCount);
			index = new LaesaBTree<T>(buffer, "data/test_pivot" + pivotCount,
					distance);
			pivotCount++;
		} else {
			Distance<T> distance = pivotList.get(pivotCount);
			index = new LaesaHeapFile<T>(buffer,
					"data/test_pivot" + pivotCount, distance);
			pivotCount++;
		}

		// insert data
		Cursor<IntKey, T> cursor = dataFile.query(new IntKey(0), new IntKey(
				Integer.MAX_VALUE));
		while (cursor.next()) {
			System.out.println("CURSOR KEY IS :" + cursor.getKey());
			index.insert(cursor.getKey(), cursor.getData());
		}

		// add index to indexlist
		indexList.add(index);
	}

	public IntKey insert(T data) throws IOException {
		dataCount++;
		IntKey key = new IntKey(dataCount);
		dataFile.insert(key, data);
		indexList.insert(key, data);
		return key;
	}

	public interface PivotCount {
		public int getPivotCount(int objectCount);
	}

	public static PivotCount pivotCountLog(final int constant, final int scale,
			final int base) {
		return new PivotCount() {
			public int getPivotCount(int objectCount) {
				return constant + scale
						* (int) (Math.log10(objectCount) / Math.log10(base));
			};
		};
	}

	public static PivotCount pivotCountConstant(final int constant) {
		return new PivotCount() {
			public int getPivotCount(int objectCount) {
				return constant;
			};
		};
	}

	public void print(Object... obj) {
		for (Object o : obj) {
			logBuffer.append(String.valueOf(o));
			System.out.print(String.valueOf(o));
		}
	}

	public void writeLog(String fileName) throws IOException {
		FileWriter writer = new FileWriter(fileName, false);
		writer.write(logBuffer.toString());
		writer.close();
		logBuffer = new StringBuffer();
		System.out.println("Written log " + fileName);
	}

	public int getPivotCount() {
		return pivotCount;
	}

	public int getDataCount() {
		return dataCount;
	}
	
	public int getDistanceCost() {
		int sum = 0;
		for (Distance<T> pivot : pivotList) {
			sum += pivot.getCalculations(); 
		}
		
		return sum;
	}
}
