package ntnustore.btree;

import static org.junit.Assert.*;

import neustore.base.DBRuntimeException;
import neustore.base.IntKey;
import neustore.base.StringData;

import org.junit.Before;
import org.junit.Test;

/**
 * JUnit test case for class {@link ntnustore.btree.SortedRecordList}.
 * 
 * @version $Id: SortedRecordListTest.java 40 2008-10-02 20:04:20Z erikbagg $
 * @author Erik Bagge Ottesen <erikbagg@idi.ntnu.no>
 */
public class SortedRecordListTest {
    SortedRecordList<IntKey, StringData> sortedList;
    public static final int TEST_KEYS[] = new int[]{Integer.MIN_VALUE, -300, -1, 0, 1, 16, 42, Integer.MAX_VALUE};

    /**
     * @throws java.lang.Exception
     */
    @Before
    public void setUp() throws Exception {
        sortedList = new SortedRecordList<IntKey, StringData>();
        
        for (int key : TEST_KEYS) {
            sortedList.insert(new IntKey(key), new StringData(255, "t" + key));
        }
    }

    /**
     * Test method for {@link ntnustore.btree.SortedRecordList#clear()}.
     */
    @Test
    public void testClear() {
        sortedList.clear();
        assertEquals(0, sortedList.size());
    }

    /**
     * Test method for {@link ntnustore.btree.SortedRecordList#size()}.
     */
    @Test
    public void testSize() {
        assertEquals(TEST_KEYS.length, sortedList.size());
    }

    /**
     * Test method for {@link ntnustore.btree.SortedRecordList#getKey(int)}.
     */
    @Test
    public void testGetKey() {
        for (int i = 0; i < TEST_KEYS.length; i++) {
            assertEquals(TEST_KEYS[i], (int)sortedList.getKey(i).value());
        }
    }

    /**
     * Test method for {@link ntnustore.btree.SortedRecordList#getData(int)}.
     */
    @Test
    public void testGetData() {
        for (int i = 0; i < TEST_KEYS.length; i++) {
            assertEquals("t" + TEST_KEYS[i], sortedList.getData(i).value());
        }
    }

    /**
     * Test method for
     * {@link ntnustore.btree.SortedRecordList#insert(neustore.base.Key, neustore.base.Data)}.
     */
    @Test
    public void testInsert() {
        IntKey key = new IntKey(-42);
        sortedList.insert(key, null);
        assertEquals(key, sortedList.getKey(2));

        key = new IntKey(Integer.MAX_VALUE - 1);
        sortedList.insert(key, null);
        assertEquals(key, sortedList.getKey(8));
        assertEquals(Integer.MAX_VALUE, (int)sortedList.getKey(9).value());
        assertEquals(TEST_KEYS.length + 2, sortedList.size());
    }
    
    /**
     * Test method for
     * {@link ntnustore.btree.SortedRecordList#insertSequential(int, neustore.base.Key, neustore.base.Data)}.
     */
    @Test
    public void testInsertSequential() {
        IntKey key = new IntKey(-42);
        sortedList.insertSequential(key, null);
        assertEquals(key, sortedList.getKey(8));

        key = new IntKey(0);
        sortedList.insertSequential(key, null);
        assertEquals(key, sortedList.getKey(9));
    }

    /**
     * Test method for
     * {@link ntnustore.btree.SortedRecordList#find(neustore.base.Key)}.
     */
    @Test
    public void testFind() {
        for (int key : TEST_KEYS) {
            assertEquals("t" + key, sortedList.find(new IntKey(key)).value());
        }
        assertNull(sortedList.find(new IntKey(-42)));
    }

    /**
     * Test method for
     * {@link ntnustore.btree.SortedRecordList#findLeq(neustore.base.Key)}.
     */
    @Test
    public void testFindLeq() {
        String lastData = null;
        for (int key : TEST_KEYS) {
            if (lastData != null) {
                assertEquals(lastData, sortedList.findLeq(new IntKey(key - 1)).value());
            }
            lastData = "t" + key;
            assertEquals(lastData, sortedList.findLeq(new IntKey(key)).value());
        }
        
        assertNotNull(sortedList.findLeq(new IntKey(-301)));
        sortedList.remove(0);
        assertNull(sortedList.findLeq(new IntKey(-301)));
    }

    /**
     * Test method for
     * {@link ntnustore.btree.SortedRecordList#findForScan(neustore.base.Key)}.
     */
    @Test
    public void testFindForScan() {
        for (int i = 0; i < TEST_KEYS.length; i++) {
            assertEquals(i - 1, sortedList.findForScan(new IntKey(TEST_KEYS[i])));
        }
        
        sortedList.clear();
        assertEquals(-1, sortedList.findForScan(new IntKey(0)));
    }

    /**
     * Test method for {@link ntnustore.btree.SortedRecordList#remove(int)}.
     */
    @Test
    public void testRemoveInt() {
        sortedList.remove(0);
        assertEquals(TEST_KEYS.length - 1, sortedList.size());
        assertEquals(TEST_KEYS[1], (int)sortedList.getKey(0).value());
        assertEquals(Integer.MAX_VALUE, (int)sortedList.getKey(sortedList.size() - 1).value());
        assertEquals(TEST_KEYS.length - 1, sortedList.size());

        while (sortedList.size() > 1) {
            sortedList.remove(sortedList.size() - 1);
        }
        assertEquals(TEST_KEYS[1], (int)sortedList.getKey(0).value());

        sortedList.remove(0);
        assertEquals(0, sortedList.size());
    }

    /**
     * Test method for
     * {@link ntnustore.btree.SortedRecordList#removeAllFrom(int)}.
     */
    @Test
    public void testRemoveAllFrom() {
        sortedList.removeAllFrom(TEST_KEYS.length - 1);
        assertEquals(TEST_KEYS.length - 1, sortedList.size());
        
        sortedList.removeAllFrom(1);
        assertEquals(TEST_KEYS[0], (int)sortedList.getKey(0).value());
        assertEquals(1, sortedList.size());
        
        sortedList.removeAllFrom(0);
        assertEquals(0, sortedList.size());
    }

    /**
     * Test method for
     * {@link ntnustore.btree.SortedRecordList#remove(neustore.base.Key)}.
     */
    @Test
    public void testRemoveKey() {
        sortedList.remove(new IntKey(TEST_KEYS[0]));
        assertNull(sortedList.find(new IntKey(TEST_KEYS[0])));
    }

}
